 /*** 
   * The contents of this file are subject to the Mozilla Public
   * License Version 1.1 (the "License"); you may not use this file
   * except in compliance with the License. You may obtain a copy of
   * the License at http://www.mozilla.org/MPL/
   *
   * Software distributed under the License is distributed on an "AS
   * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   * implied. See the License for the specific language governing
   * rights and limitations under the License.
   *
   * The Original Code is this file as it was released on
   * June 7, 2002.
   *
   * The Initial Developer of the Original Code is HJ van Rantwijk.
   * Portions created by HJ van Rantwijk are Copyright (C) 2002-2005
   * HJ van Rantwijk.  All Rights Reserved.
   *
   */

const kDebug               = false;
const kUpdateCheckDelay    = 180000; // 1 * 60 * 1000; // 3 * 60 * 1000 = 3 minutes
const kUNEnabledPref       = "multizilla.update_notifications.enabled";
const kUNDatasourceURIPref = "multizilla.update_notifications.provider.datasource"; // http://multizilla.mozdev.org/xpi/updates.rdf
const kUNFrequencyPref     = "multizilla.update_notifications.provider.frequency";
const kUNLastCheckedPref   = "multizilla.update_notifications.provider.last_checked";
const kUNBundleURI         = "chrome://multiviews/locale/update-notifications.properties";

////////////////////////////////////////////////////////////////////////
//   mzUpdateNotifier : nsIProfileStartupListener, nsIObserver
////////////////////////////////////////////////////////////////////////
var mzUpdateNotifier = 
{
  onProfileStartup: function(aProfileName)
  {
    debug("onProfileStartup");
    var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
    observerService.addObserver(this, "domwindowopened", false);
  },

  mTimer: null, // need to hold on to timer ref

  observe: function(aSubject, aTopic, aData)
  {
    debug("observe: " + aTopic);
    if (aTopic == "domwindowopened") {
      try {
        const kITimer = Components.interfaces.nsITimer;
        this.mTimer = Components.classes["@mozilla.org/timer;1"].createInstance(kITimer);
        this.mTimer.init(this, kUpdateCheckDelay, kITimer.TYPE_ONE_SHOT);
        // we are no longer interested in the ``domwindowopened'' topic
        var observerService = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
        observerService.removeObserver(this, "domwindowopened");
        // remove our timer reference on XPCOM shutdown to avoid a leak.
        observerService.addObserver(this, "xpcom-shutdown", false);
      }
      catch (ex) {
        debug("Exception init'ing timer: " + ex);
      }
    }
    else if (aTopic == "timer-callback") {
      this.mTimer = null; // free up timer so it can be gc'ed
      this.checkForUpdate();
    }
    else if (aTopic == "xpcom-shutdown") {
      /* 
       * We need to drop our timer reference here to avoid a leak
       * since the timer keeps a reference to the observer. 
       */
      this.mTimer = null;
    }
  },

  checkForUpdate: function()
  {
    debug("checkForUpdate");
    if (this.shouldCheckForUpdate()) {
      try {
        var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
        var updateDatasourceURI = prefs.getComplexValue(kUNDatasourceURIPref,Components.interfaces.nsIPrefLocalizedString).data;
        var rdf = Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
        var ds = rdf.GetDataSource(updateDatasourceURI);
        ds = ds.QueryInterface(Components.interfaces.nsIRDFXMLSink);
        ds.addXMLSinkObserver(mzUpdateDatasourceObserver);
      }
      catch (ex) {
        debug("Exception getting updates.rdf: " + ex);
      }
    }
  },

  shouldCheckForUpdate: function()
  {
    debug("shouldCheckForUpdate");
    return true; // For testing purposes only!
    var shouldCheck = false;
    try {
      var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
      if (prefs.getBoolPref(kUNEnabledPref)) {
        var freq = prefs.getIntPref(kUNFrequencyPref) * 86400; // (24 * 60 * 60); // secs
        var now = (new Date().valueOf())/1000; // secs
        if (!prefs.prefHasUserValue(kUNLastCheckedPref)) {
          var randomizedLastChecked = now + freq * (1 + Math.random());
          prefs.setIntPref(kUNLastCheckedPref, randomizedLastChecked);
          return false;
        }
        var lastChecked = prefs.getIntPref(kUNLastCheckedPref);
        if ((lastChecked + freq) > now)
          return false;
        prefs.setIntPref(kUNLastCheckedPref, now);
        prefs = prefs.QueryInterface(Components.interfaces.nsIPrefService);
        prefs.savePrefFile(null);
        shouldCheck = true;
      }
    }
    catch (ex) {
      shouldCheck = false;
      debug("Exception in shouldCheckForUpdate: " + ex);
    }
    return shouldCheck;
  },

  QueryInterface: function(aIID)
  {
    if (aIID.equals(Components.interfaces.nsIObserver) ||
        aIID.equals(Components.interfaces.nsIProfileStartupListener) ||
        aIID.equals(Components.interfaces.nsISupports))
          return this;
    Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
    return null;
  }
}
////////////////////////////////////////////////////////////////////////
//   mzUpdateDatasourceObserver : nsIRDFXMLSinkObserver
////////////////////////////////////////////////////////////////////////
var mzUpdateDatasourceObserver = 
{
  onBeginLoad: function(aSink) {},
  onInterrupt: function(aSink) {},
  onResume: function(aSink)    {},

  onEndLoad: function(aSink)
  {
    debug("onEndLoad");
    aSink.removeXMLSinkObserver(this);
    var dataSource = aSink.QueryInterface(Components.interfaces.nsIRDFDataSource);
    var productName = this.getProductName();
    dump("\nproductName: " + productName + "\n");
    var updateInfo = this.getUpdateInfo(dataSource, productName);
    var productObj = { name    : updateInfo.productName,
                       version : updateInfo.version,
                       revision: updateInfo.revision };
    var updateAvailable = this.newerVersionAvailable(productObj);

    if (updateInfo && updateAvailable) {
      debug("Update for: " + updateInfo.productName + " available!");
      var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
      var winWatcher = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].getService(Components.interfaces.nsIWindowWatcher);
      var unBundle = this.getBundle(kUNBundleURI);

      if (!unBundle)
        return;

      var title = unBundle.formatStringFromName("updateDialogTitle", [updateInfo.productName], 1);
      var description = unBundle.formatStringFromName("updateDialogText", [updateInfo.productName, productObj.version, productObj.revision, updateInfo.productName, updateInfo.version, updateInfo.revision, updateInfo.productName], 7);
      var button0Text = unBundle.GetStringFromName("getItNow");
      var button1Text = unBundle.GetStringFromName("noThanks");
      var checkMsg = unBundle.GetStringFromName("dontAskAgain");
      var checkVal = {value:0};
      var result = promptService.confirmEx(winWatcher.activeWindow, title, description, 
                                           (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_IS_STRING) +
                                           (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_IS_STRING),
                                           button0Text, button1Text, null, checkMsg, checkVal);

      if (result == 0) {
        var ass = Components.classes["@mozilla.org/appshell/appShellService;1"].getService(Components.interfaces.nsIAppShellService);
        var installationText = updateInfo.text;
        var installationURL = updateInfo.URL;
        ass.hiddenDOMWindow.InstallTrigger.install({installationText : installationURL});
      }
      if (checkVal.value) {
        var prefs = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
        prefs.setBoolPref(kUNEnabledPref, false);
      }
    }
    else {
      debug("No update for: " + updateInfo.productName + " available!");
    }
  },

  onError: function(aSink, aStatus, aErrorMsg)
  {
    debug("Error " + aStatus + ": " + aErrorMsg);
    aSink.removeXMLSinkObserver(this);
  },

  getBrowserWindow: function()
  {
    var windowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].getService(Components.interfaces.nsIWindowMediator);
    return windowManager.getMostRecentWindow("navigator:browser");
  },

  getBundle: function(aURI)
  {
    if (!aURI)
      return null;
    var bundle = null;
    try {
      var strBundleService = Components.classes["@mozilla.org/intl/stringbundle;1"].getService(Components.interfaces.nsIStringBundleService);
      bundle = strBundleService.createBundle(aURI);
    }
    catch (ex) {
      bundle = null;
      debug("Exception getting bundle " + aURI + ": " + ex);
    }
    return bundle;
  },

  getProductName: function()
  {
    var browserWindow = this.getBrowserWindow();
    var tabbrowser = (browserWindow) ? browserWindow.getBrowser() : null;

    return (tabbrowser && 'mProduct' in tabbrowser) ? tabbrowser.mProduct : undefined;
  },

  getUpdateInfo: function(aDS, aProductName)
  {
    var updateInfo = new Object;

    try {
      var rdf = Components.classes["@mozilla.org/rdf/rdf-service;1"].getService(Components.interfaces.nsIRDFService);
      var src = "urn:updates:latest";
      updateInfo.registryName = this.getTarget(rdf, aDS, src, "registryName");
      updateInfo.productName  = this.getTarget(rdf, aDS, src, "productName");

      if (aProductName == undefined || aProductName == "undetermined")
        aProductName = "Win";
      else
        aProductName += "-";

      updateInfo.version      = this.getTarget(rdf, aDS, src, aProductName + "Version");
      updateInfo.revision     = this.getTarget(rdf, aDS, src, aProductName + "Revision");
      updateInfo.text         = this.getTarget(rdf, aDS, src, aProductName + "Text");
      updateInfo.URL          = this.getTarget(rdf, aDS, src, aProductName + "URL");
    }
    catch (ex) {
      debug("Exception getting update info: " + ex);
      updateInfo = null;
    }
    return updateInfo;
  },

  getTarget: function(aRDF, aDS, aSrc, aProp)
  {
    var src = aRDF.GetResource(aSrc);
    var arc = aRDF.GetResource("http://home.netscape.com/NC-rdf#" + aProp);
    var target = aDS.GetTarget(src, arc, true);
    return target.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
  },

  newerVersionAvailable: function(aProductObj)
  {
    var browserWindow = this.getBrowserWindow();

    if (!browserWindow)
      return false;

    if (aProductObj.name == "MultiZilla") {
      var mzVersion = browserWindow.gBrowser.mVersion;
      var mzRevision = browserWindow.gBrowser.mVersionRev;

      if (mzVersion < aProductObj.version || (mzVersion == aProductObj.version && mzRevision < aProductObj.revision)) {
        aProductObj.version = mzVersion;
        aProductObj.revision = mzRevision;
        return true;
      }
    }
    else if (aProductObj.name == "GoogleBox") {
      var mzToolbox = browserWindow.document.getElementById("MultiZillaToolbar");

      if (mzToolbox.mToolboxActivated) { // GoogleBox installed?
        var googleBox = browserWindow.document.getElementById("internalToolBox");
        // Pre GoogleBox v1.4a versions had no version info, that's why we check for 'undefined'
        var googleBoxVersion = (googleBox.mVersion == undefined) ? 0 : googleBox.mVersion;
        var googleBoxRevision = (googleBox.mVersionRev == undefined) ? "" : googleBox.mVersionRev;

        if (googleBoxVersion < aVersion || googleBoxRevision < aRevision) {
          aProductObj.version = googleBoxVersion;
          aProductObj.revision = googleBoxRevision;
          return true;
        }
      }
      return false;
    }
  }
}
////////////////////////////////////////////////////////////////////////
//  mzUpdateNotifierModule : nsIModule
////////////////////////////////////////////////////////////////////////
var mzUpdateNotifierModule = 
{
  mClassName:     "Update Notifier",
  mContractID:    "@mozilla.org/update-notifier;1",
  mClassID:       Components.ID("8b6dcf5e-3b5a-4fff-bff5-65a8fa9d71b2"),

  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aCID.equals(this.mClassID))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    return this.mFactory;
  },

  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    if (kDebug)
      dump("*** Registering: mzUpdateNotifier ***\n");
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(this.mClassID, this.mClassName, this.mContractID, aFileSpec, aLocation, aType);
    this.getCategoryManager().addCategoryEntry("profile-startup-category", this.mContractID, "", true, true);
  },

  unregisterSelf: function(aCompMgr, aFileSpec, aLocation)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(this.mClassID, aFileSpec);
    this.getCategoryManager().deleteCategoryEntry("profile-startup-category", this.mContractID, true);
  },

  canUnload: function(aCompMgr)
  {
    return true;
  },

  getCategoryManager: function()
  {
    return Components.classes["@mozilla.org/categorymanager;1"].
      getService(Components.interfaces.nsICategoryManager);
  },
  //////////////////////////////////////////////////////////////////////
  //   mFactory : nsIFactory
  //////////////////////////////////////////////////////////////////////
  mFactory:
  {
    createInstance: function(aOuter, aIID)
    {
      if (aOuter != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;
      if (!aIID.equals(Components.interfaces.nsIObserver) &&
          !aIID.equals(Components.interfaces.nsIProfileStartupListener) &&
          !aIID.equals(Components.interfaces.nsISupports))
        throw Components.results.NS_ERROR_INVALID_ARG;

      return mzUpdateNotifier.QueryInterface(aIID);
    },

    lockFactory: function(aLock)
    {
      // quiten warnings
    }
  }
};

function NSGetModule(aCompMgr, aFileSpec)
{
  return mzUpdateNotifierModule;
}

////////////////////////////////////////////////////////////////////////
//   Debug helper
////////////////////////////////////////////////////////////////////////
if (!kDebug)
  debug = function(m) {};
else
  debug = function(m) {dump("*** mzUpdateNotifier: " + m + "\n");};
